/********************************************************************************/
/* 	fmfloader.d																	*/
/*------------------------------------------------------------------------------*/
/*	製作		kenmo（あるふぁ～秘密基地）										*/
/*	製作開始	2008/05/22														*/
/*	MAIL		dk@red.interq.or.jp												*/
/*	URL			http://maglog.jp/alpha-secret-base/								*/
/*																				*/
/*	このソースは「やわらかライセンス」の元で配布されています。					*/
/*-更新履歴---------------------------------------------------------------------*/
/*	2008/05/22	初版															*/
/*	2008/05/24	Layer2Dにchip_widthとchip_heightを追加							*/
/*-その他-----------------------------------------------------------------------*/
/*	なし																		*/
/********************************************************************************/
import std.file;

/**
 * 2次元マップ
 */
class Layer2D
{
public:
	int[] data;   // データテーブル
	int   width;  // 幅
	int   height; // 高さ
	int   chip_width;  // チップの幅
	int   chip_height; // チップの高さ
public:
	this(int w, int h)
	{
		this(w, h, new char[w*h]);
	}
	/**
	 * コンストラクタ
	 * @param w  幅
	 * @param h 高さ
	 * @param data   データ
	 */
	this(int w, int h, char[] data)
	{
		this.width  = w;
		this.height = h;
		this.data   = new int[width*height];
		for(int i = 0; i < width*height; i++)
		{
			this.data[i] = cast(int)data[i];
		}
	}
	/**
	 * 座標を指定して値を取得
	 * @param x, y 座標
	 */
	int get(int x, int y)
	{
		if(x < 0 || width <= x || y < 0 || height <= y) return -1; // 場合によっては例外吐いてもいいかも
		return data[y*width+x];
	}
	/**
	 * 座標を指定して値を設定
	 * @param x, y 座標
	 * @param val  値
	 */
	void set(int x, int y, int val)
	{
		if(x < 0 || width <= x || y < 0 || height <= y) return;
		data[y*width+x] = val;
	}
	/**
	 * 指定の値で全てを埋める
	 */
	void fill(int val)
	{
		for(int i = 0; i < width*height; i++) data[i] = val;
	}
	/**
	 * レイヤーの値をコピー
	 * @param layer コピー元レイヤー
	 */
	void copy(Layer2D layer)
	{
		if(width != layer.width || height != layer.height) return;
		
		for(int i = 0; i < width*height; i++)
		{
			data[i] = layer.data[i];
		}
	}
	/**
	 * デバッグ出力
	 */
	void dump()
	{
		for(int j = 0; j < height; j++)
		{
			for(int i = 0; i < width; i++)
			{
				printf("%d", get(i, j));
			}
			printf("\n");
		}
	}
}

/**
 * FMF読み込みクラス
 */
class FMFLoader
{
public:
	char[]    identifier; // 識別文字「FMF_」
	int       datasize;   // ヘッダを除いたデータサイズ
	int       width;      // マップの幅
	int       height;     // マップの高さ
	int       chipwidth;  // チップの幅
	int       chipheight; // チップの高さ
	int       layercount; // レイヤー数
	int       bitcount;   // データのbit数(8/16)
	Layer2D[] layerArray; // レイヤー配列
private:
	char[] _file;
	int    _ptr;
public:
	/**
	 * コンストラクタ<br>
	 * TODO: マップデータは8bitのみ
	 * @param filepath FMFファイル
	 * @throw ファイルフォーマット例外
	 */
	this(char[] filepath)
	{
		_file = cast(char[])read(filepath);
		_ptr  = 0;
		identifier = _file.dup[0..4]; // ファイル識別子  [4byte]
		if(identifier != "FMF_")
		{
			throw new Error("Invalid FMF format (" ~ filepath ~ ")");
		}
		_ptr += 4;
		datasize   = readLong(); // ヘッダを除いたサイズ [4byte]
		width      = readLong(); // マップの幅           [4byte]
		height     = readLong(); // マップの高さ         [4byte]
		chipwidth  = readByte(); // チップの幅           [1byte]
		chipheight = readByte(); // チップの高さ         [1byte]
		layercount = readByte(); // レイヤー数           [1byte]
		bitcount   = readByte(); // データのbit数(8/16)  [1byte]
		layerArray = new Layer2D[layercount];
		for(int i = 0; i < layercount; i++)
		{
			Layer2D layer = new Layer2D(width, height, _file[_ptr.._ptr+width*height]);
			layerArray[i] = layer;
			layerArray[i].chip_width = chipwidth;
			layerArray[i].chip_height = chipheight;
			_ptr += width*height;
		}
	}
	/**
	 * 指定レイヤーを取得する
	 * @param id レイヤーID
	 * @return 2次元マップ
	 */
	Layer2D getLayer(int id)
	{
		if(id < 0 || id >= layercount) return null;
		return layerArray[id];
	}
	/**
	 * デバッグ出力
	 */
	void dump()
	{
		printf("identifier='%.*s' datasize=%d (w,h)=(%d,%d) chip(x,y)=(%d,%d) layer=%d bit=%d\n",
			identifier, datasize, width, height, chipwidth, chipheight, layercount, bitcount);
		foreach(i, layer; layerArray)
		{
			printf("Layer-%d\n", i);
			layer.dump();
		}
	}

private:
	/**
	 * 1byte読み込み
	 */
	int readByte()
	{
		if(_ptr - 1 >= cast(int)_file.length) return -1;
		int result = _file[_ptr];
		_ptr += 1;
		return result;
	}
	/**
	 * 4byte読み込み
	 */
	int readLong()
	{
		if(_ptr - 4 >= cast(int)_file.length) return -1;
		int result = _file[_ptr] + (_file[_ptr+1] << 8) + (_file[_ptr+2] << 16) + (_file[_ptr+3] << 24);;
		_ptr += 4;
		return result;
	}
}
